/*
 * broker_allup.cc
 *
 * Copyright (c) 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file broker_allup.cc
 */

#include "config.h"

#include <stdlib.h>
#include <signal.h>

#include <sys/time.h>

#include <assert.h>
#include <assert_pp.h>

#include <timesFile.h>

#include <rk/rk.h>
#include <rk/rk_error.h>

#include <iostream>

#include <BrokerC.h>
#include <StrictPolicyC.h>

#include <ManagerImpl.hh>

#include <StrictPolicyImpl.hh>

#include <ExactTaskFactory.hh>
#include <MaxTaskFactory.hh>
#include <MaxDecayTaskFactory.hh>
#include <GlacialTaskFactory.hh>
#include <GCTaskFactory.hh>

#include <BasicTaskKernelImpl.h>
#include <BasicTaskQosketImpl.hh>
#include <ExactTaskFactory_Quo.hh>

static void write_ior(CORBA::ORB_ptr orb,
		      CORBA::Object_ptr obj,
		      const char *to)
{
    require(to != NULL);
    
    ofstream ostr(to);

    ostr << orb->object_to_string(obj) << endl;
    ostr.close();
}

static struct ServerMainData {
    unsigned long smd_RunTime;
#if defined(HAVE_LIBRK)
    rk_resource_set_t smd_ResourceSet;
#endif
} server_main_data;

static void sig_exit_handler(int sig)
{
#if defined(HAVE_LIBRK)
    if( server_main_data.smd_ResourceSet != NULL )
    {
	rk_resource_set_destroy(server_main_data.smd_ResourceSet);
    }
#endif
    exit(EXIT_SUCCESS);
}

int main(int argc, char *argv[])
{
    int lpc, retval = EXIT_FAILURE;
    
    signal(SIGINT, sig_exit_handler);
    signal(SIGTERM, sig_exit_handler);
    
    for( lpc = 1; lpc < argc; lpc += 2 )
    {
	if( strcasecmp(argv[lpc], "-for") == 0 )
	{
	    if( sscanf(argv[lpc + 1],
		       "%lu",
		       &server_main_data.smd_RunTime) == 0 )
	    {
		fprintf(stderr,
			"'-for' value is not an integer: %s\n",
			argv[lpc + 1]);
	    }
	}
	else
	{
	    fprintf(stderr, "Unknown flag: %s\n", argv[lpc]);
	}
    }

    if( server_main_data.smd_RunTime )
    {
	struct itimerval itv;
	
	signal(SIGALRM, sig_exit_handler);
	itv.it_interval.tv_sec = 0;
	itv.it_interval.tv_usec = 0;
	itv.it_value.tv_sec = server_main_data.smd_RunTime;
	itv.it_value.tv_usec = 0;
	if( setitimer(ITIMER_REAL, &itv, NULL) < 0 )
	{
	    perror("setitimer");
	}
    }
    
#if defined(HAVE_ACE)
    /* Put a sock in ACE */
    ACE_Log_Msg::instance()->priority_mask(0, ACE_Log_Msg::PROCESS);
#endif

    /* ORB/POA setup */
    CORBA::ORB_var orb = CORBA::ORB_init(argc, argv);
    CORBA::Object_var obj = orb->resolve_initial_references("RootPOA");
    
    PortableServer::POA_var root_poa = PortableServer::POA::_narrow(obj.in());
    PortableServer::POAManager_var mgr = root_poa->the_POAManager();
    mgr->activate();

    PortableServer::ThreadPolicy_var thread = root_poa->
	create_thread_policy(PortableServer::SINGLE_THREAD_MODEL);
    
    CORBA::PolicyList policy_list;
    policy_list.length(1);
    policy_list[0] = PortableServer::ThreadPolicy::_duplicate(thread.in());
    
    PortableServer::POA_var st_poa = root_poa->
	create_POA("SingleThread",
		   PortableServer::POAManager::_nil(),
		   policy_list);
    
    thread->destroy();

    quo::QuoKernel_var qk;
    
    Qosket::Basic::BasicTaskKernelImpl *btki;

    /* ... no joy on Java kernel, create our own. */
    btki = new Qosket::Basic::BasicTaskKernelImpl(orb);
    qk = btki->_this();

    write_ior(orb.in(), qk.in(), "quoKernel.ior");
    
    /* Broker setup */
    Broker::Manager_var manager;
    ManagerImpl *mi;

    mi = new ManagerImpl("Manager");
    manager = mi->_this();

    write_ior(orb.in(), manager.in(), "manager.ior");

    StrictPolicy_var policy;
    StrictPolicyImpl *spi;

    spi = new StrictPolicyImpl("StrictPolicy");
    policy = spi->_this();

    write_ior(orb.in(), policy.in(), "strict_policy.ior");

    mi->CurrentPolicy(policy.in());

    Broker::TaskFactory_var tf;

    ExactTaskFactory *etf;

    etf = new ExactTaskFactory("ExactTaskFactory");
    tf = etf->_this();

    write_ior(orb.in(), tf.in(), "exact_task_factory.ior");

    MaxTaskFactory *mtf;

    mtf = new MaxTaskFactory("MaxTaskFactory");
    tf = mtf->_this();

    write_ior(orb.in(), tf.in(), "max_task_factory.ior");

    MaxDecayTaskFactory *mdtf;

    mdtf = new MaxDecayTaskFactory("MaxDecayTaskFactory");
    tf = mdtf->_this();

    write_ior(orb.in(), tf.in(), "max_decay_task_factory.ior");

    GlacialTaskFactory *gtf;

    gtf = new GlacialTaskFactory("GlacialTaskFactory");
    tf = gtf->_this();

    write_ior(orb.in(), tf.in(), "glacial_task_factory.ior");

    GCTaskFactory *gctf;

    gctf = new GCTaskFactory("GCTaskFactory");
    tf = gctf->_this();

    write_ior(orb.in(), tf.in(), "gc_task_factory.ior");

#if defined(HAVE_LIBRK)
    {
	rk_resource_set_t rs;
	
	if( (rs = rk_resource_set_create("broker_allup")) != NULL )
	{
	    struct cpu_reserve_attr cra;
	    rk_reserve_t cr;
	    
	    memset(&cra, 0, sizeof(cra));
	    cra.compute_time.tv_sec = 0;
	    cra.compute_time.tv_nsec = 10000000;
	    cra.period.tv_sec = 1;
	    cra.period.tv_nsec = 0;
	    cra.deadline = cra.period;
	    cra.reserve_type.sch_mode = RSV_SOFT;
	    cra.reserve_type.enf_mode = RSV_SOFT;
	    cra.reserve_type.rep_mode = RSV_SOFT;
	    cra.processor = RK_ANY_CPU;
	    if( (rk_cpu_reserve_create(rs, &cr, &cra) ==
		 RK_SUCCESS) &&
		(rk_resource_set_attach_process(rs, getpid()) ==
		 RK_SUCCESS) )
	    {
		server_main_data.smd_ResourceSet = rs;
	    }
	    else
	    {
		rk_resource_set_destroy(rs);
	    }
	}
	else
	{
	    perror("rk_resource_set_create");
	}
    }
#endif
    
    orb->run();

    orb->shutdown();

#if defined(HAVE_LIBRK)
    if( server_main_data.smd_ResourceSet != NULL )
    {
	rk_resource_set_destroy(server_main_data.smd_ResourceSet);
    }
#endif
    
    return( retval );
}
