/*
 * RKTask.cc
 *
 * Copyright (c) 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file RKTask.cc
 *
 * Implementation of the RKTask class.
 */

#include "config.h"

#include <string.h>

#include <iostream>

#include <assert_pp.h>

#include "RKTask.hh"

static void microsec_to_timespec (CORBA::ULong microseconds,
				  struct timespec &time_spec)
{
  time_spec.tv_sec =  (int)(microseconds / 1000000);
  time_spec.tv_nsec = (int)(microseconds % 1000000)*1000;
}

static void timespec_to_microsec (struct timespec time_spec,
				  CORBA::ULong &microseconds)
{
  microseconds = time_spec.tv_sec * 1000000;
  microseconds += time_spec.tv_nsec / 1000;
}

RKTask::RKTask(const Broker::TaskParameters &tp)
    throw (CORBA::SystemException,
	   Broker::DuplicateTaskParameter,
	   Broker::InvalidTaskParameter,
	   Broker::MissingTaskParameter)
{
    unsigned int lpc;
    pid_t pid = -1;

    gettimeofday(&this->rkt_StartTime, NULL);
    this->rkt_LastReservationLog = this->rkt_StartTime;
    this->rkt_ReservationLog = NULL;
    
    memset(&this->rkt_CPUReserveSpec, 0, sizeof(this->rkt_CPUReserveSpec));
    microsec_to_timespec(100, this->rkt_CPUReserveSpec.compute_time);
    this->rkt_CPUReserve = NULL_RESERVE;

    /* Process the arguments */
    for( lpc = 0; lpc < tp.length(); lpc++ )
    {
	if( strcasecmp(tp[lpc].name.in(), "name") == 0 )
	{
	    const char *name;

	    if( this->rkt_Name != NULL )
	    {
		throw Broker::DuplicateTaskParameter("name");
	    }
	    else if( tp[lpc].value >>= name )
	    {
		this->rkt_Name = name;
	    }
	    else
	    {
		throw Broker::InvalidTaskParameter("'name' not a string",
						   tp[lpc]);
	    }
	}
	else if( strcasecmp(tp[lpc].name.in(), "pid") == 0 )
	{
	    CORBA::Long c_pid;

	    if( pid != -1 )
	    {
		throw Broker::DuplicateTaskParameter("pid");
	    }
	    else if( tp[lpc].value >>= c_pid )
	    {
		pid = c_pid;
	    }
	    else
	    {
		throw Broker::InvalidTaskParameter("'pid' not a long",
						   tp[lpc]);
	    }
	}
	else if( (strcasecmp(tp[lpc].name.in(), "res_log") == 0) ||
		 (strcasecmp(tp[lpc].name.in(), "res-log") == 0) ||
		 (strcasecmp(tp[lpc].name.in(), "res log") == 0) )
	{
	    const char *res_log;
	    
	    if( this->rkt_ReservationLog != NULL )
	    {
		throw Broker::DuplicateTaskParameter("res log");
	    }
	    else if( tp[lpc].value >>= res_log )
	    {
		this->rkt_ReservationLog = new ofstream(res_log);
		(*this->rkt_ReservationLog) << setfill('0');
		(*this->rkt_ReservationLog)
		    << "# start: "
		    << (long)this->rkt_StartTime.tv_sec
		    << "."
		    << setw(6) << (long)this->rkt_StartTime.tv_usec
		    << endl;
	    }
	    else
	    {
		throw Broker::InvalidTaskParameter("'res log' not a string",
						   tp[lpc]);
	    }
	}
    }
    
    /* Make sure we have everything we need. */
    if( this->rkt_Name == NULL )
    {
	throw Broker::MissingTaskParameter("name");
    }
    if( pid == -1 )
    {
	throw Broker::MissingTaskParameter("pid");
    }

    /* Go ahead and create a resource set for the process. */
    if( (this->rkt_ResourceSet = rk_proc_get_rset(pid)) != NULL )
    {
	/* We'll just manage the existing resource set. */
    }
    else if( (this->rkt_ResourceSet =
	      rk_resource_set_create(this->rkt_Name.in())) != NULL )
    {
	if( rk_resource_set_attach_process(this->rkt_ResourceSet, pid) == 0 )
	{
	    /*
	     * Joy, we'll attach a CPU reservation later in BeginCPUScheduling
	     */
	}
	else if( errno != EBUSY )
	{
	    rk_resource_set_destroy(this->rkt_ResourceSet);
	    this->rkt_ResourceSet = NULL_RESOURCE_SET;
	    throw CORBA::NO_MEMORY(); // XXX
	}
    }
    else
    {
	throw CORBA::NO_MEMORY(); // XXX
    }
}

RKTask::~RKTask()
{
    rk_resource_set_destroy(this->rkt_ResourceSet);
    this->rkt_ResourceSet = NULL_RESOURCE_SET;
    this->rkt_CPUReserve = NULL_RESERVE;
    if( this->rkt_ReservationLog != NULL )
	delete this->rkt_ReservationLog;
}

char *RKTask::Name(void)
    throw (CORBA::SystemException)
{
    CORBA::String_var retval;

    retval = this->rkt_Name;
    return( retval._retn() );
}

CORBA::ULong RKTask::Period(void)
    throw (CORBA::SystemException)
{
    CORBA::ULong retval;

    timespec_to_microsec(this->rkt_CPUReserveSpec.period, retval);
    return( retval );
}

CORBA::ULong RKTask::Deadline(void)
    throw (CORBA::SystemException)
{
    CORBA::ULong retval;

    timespec_to_microsec(this->rkt_CPUReserveSpec.deadline, retval);
    return( retval );
}

void RKTask::BeginCPUScheduling(Broker::Manager_ptr manager,
				const Broker::ScheduleParameters &cs)
    throw (CORBA::SystemException,
	   Broker::DuplicateScheduleParameter,
	   Broker::InvalidScheduleParameter,
	   Broker::MissingScheduleParameter)
{
    CORBA::ULong start = 0, period = ~0, deadline = ~0;
    unsigned int lpc;
    
    if( this->rkt_CPUReserve != NULL_RESERVE )
    {
	throw CORBA::BAD_INV_ORDER();
    }
    
    /* Process the arguments */
    if( CORBA::is_nil(manager) )
    {
	throw CORBA::BAD_PARAM();
    }
    for( lpc = 0; lpc < cs.length(); lpc++ )
    {
	if( strcasecmp(cs[lpc].name.in(), "start") == 0 )
	{
	    /* XXX wrong type */
	    if( !(cs[lpc].value >>= start) )
	    {
		throw Broker::InvalidScheduleParameter("'start' not a ulong",
						       cs[lpc]);
	    }
	}
	else if( strcasecmp(cs[lpc].name.in(), "period") == 0 )
	{
	    if( period != ~0UL )
	    {
		throw Broker::DuplicateScheduleParameter("period");
	    }
	    else if( !(cs[lpc].value >>= period) )
	    {
		throw Broker::InvalidScheduleParameter("'period' not a ulong",
						       cs[lpc]);
	    }
	}
	else if( strcasecmp(cs[lpc].name.in(), "deadline") == 0 )
	{
	    if( deadline != ~0UL )
	    {
		throw Broker::DuplicateScheduleParameter("deadline");
	    }
	    else if( !(cs[lpc].value >>= deadline) )
	    {
		throw Broker::InvalidScheduleParameter(
			"'deadline' not a ulong",
			cs[lpc]);
	    }
	}
    }
    
    /* Make sure we have everything we need */
    if( period == ~0UL )
    {
	throw Broker::MissingScheduleParameter("period");
    }

    /* ... and derive the rest. */
    if( deadline == ~0UL )
    {
	deadline = period;
    }

    this->rkt_Manager = Broker::Manager::_duplicate(manager);

    /* XXX Do more checking here... */
    if( (this->rkt_CPUReserveSpec.compute_time.tv_sec == 0) &&
	(this->rkt_CPUReserveSpec.compute_time.tv_nsec == 0) )
    {
	microsec_to_timespec(100, this->rkt_CPUReserveSpec.compute_time);
    }
    microsec_to_timespec(period, this->rkt_CPUReserveSpec.period);
    microsec_to_timespec(deadline, this->rkt_CPUReserveSpec.deadline);
    this->rkt_CPUReserveSpec.reserve_type.sch_mode = RSV_SOFT;
    this->rkt_CPUReserveSpec.reserve_type.enf_mode = RSV_SOFT;
    this->rkt_CPUReserveSpec.reserve_type.rep_mode = RSV_SOFT;
    this->rkt_CPUReserveSpec.processor = RK_ANY_CPU;

    manager->AddTask(this->_this());
    
    switch( rk_cpu_reserve_create(this->rkt_ResourceSet,
				  &this->rkt_CPUReserve,
				  &this->rkt_CPUReserveSpec) )
    {
    case 0:
	break;
    case EINVAL:
	throw Broker::InvalidScheduleParameter();
    default:
	throw Broker::Internal("Unhandled return value at: "
			       __FILE__
			       ":"
			       __STRING(__LINE__));
    }
}

void RKTask::EndCPUScheduling(void)
    throw (CORBA::SystemException)
{
    if( this->rkt_CPUReserve == NULL_RESERVE )
    {
	throw CORBA::BAD_INV_ORDER();
    }

    rk_cpu_reserve_delete(this->rkt_ResourceSet);
    this->rkt_CPUReserve = NULL_RESERVE;
    this->rkt_Manager->RemoveTask(this->_this());

    this->rkt_Manager = Broker::Manager::_nil();
}

CORBA::ULong RKTask::GetComputeTime(void)
    throw (CORBA::SystemException)
{
    CORBA::ULong retval;

    timespec_to_microsec(this->rkt_CPUReserveSpec.compute_time, retval);
    return( retval );
}

void RKTask::SetComputeTime(CORBA::ULong usecs)
    throw (CORBA::SystemException)
{
    struct timespec old_ct;

    old_ct = this->rkt_CPUReserveSpec.compute_time;
    microsec_to_timespec(usecs, this->rkt_CPUReserveSpec.compute_time);

    if( this->rkt_CPUReserve != NULL_RESERVE )
    {
	switch( rk_cpu_reserve_ctl(this->rkt_ResourceSet,
				   &this->rkt_CPUReserveSpec) )
	{
	case 0:
	    if( this->rkt_ReservationLog != NULL )
	    {
		struct timeval tv, st, diff;
		
		timersub(&this->rkt_LastReservationLog,
			 &this->rkt_StartTime,
			 &diff);
		st.tv_sec = usecs / 1000000;
		st.tv_usec = usecs % 1000000;
		gettimeofday(&tv, NULL);
		timersub(&tv, &this->rkt_StartTime, &diff);
		(*this->rkt_ReservationLog)
		    << (long)diff.tv_sec
		    << "."
		    << setw(6) << (long)diff.tv_usec
		    << " "
		    << (long)st.tv_sec
		    << "."
		    << setw(6) << (long)st.tv_usec
		    << endl;
		this->rkt_LastReservationLog = tv;
	    }
	    break;
	default:
	    switch( errno )
	    {
	    case EINVAL:
		ensure(0);
		break;
	    case ENODEV:
		ensure(0);
		break;
	    case ENOSPC:
		cerr << "** SetComputeTime failed! "
		     << usecs
		     << endl;
		break;
	    case EFAULT:
		ensure(0);
		break;
	    case EACCES:
		cerr << "** No permission to change compute time? "
		     << usecs
		     << endl;
		break;
	    case ENOSYS:
		ensure(0);
		break;
	    }
	    this->rkt_CPUReserveSpec.compute_time = old_ct;
	    throw CORBA::BAD_PARAM();
	}
    }
    else
    {
	/* We're not scheduling at the moment, just remember the value. */
	if( this->rkt_ReservationLog != NULL )
	{
	    struct timeval tv, st, diff;
	    
	    timersub(&this->rkt_LastReservationLog,
		     &this->rkt_StartTime,
		     &diff);
	    st.tv_sec = usecs / 1000000;
	    st.tv_usec = usecs % 1000000;
	    gettimeofday(&tv, NULL);
	    timersub(&tv, &this->rkt_StartTime, &diff);
	    (*this->rkt_ReservationLog)
		<< (long)diff.tv_sec
		<< "."
		<< setw(6) << (long)diff.tv_usec
		<< " "
		<< (long)st.tv_sec
		<< "."
		<< setw(6) << (long)st.tv_usec
		<< endl;
	    this->rkt_LastReservationLog = tv;
	}
    }
}
