/*
 * HeyParser.cc
 *
 * Copyright (c) 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file HeyParser.cc
 *
 * Implementation file for the HeyParser class.
 */

#include "config.h"

#include <string.h>
#include <assert_pp.h>

#include "HeyParser.hh"

HeyParser::HeyParser(int argc, const char *argv[])
    throw (HeyParserException)
{
    require(argc > 0);
    require(argv != NULL);

    /* Need atleast three args (program name, server name, what value) */
    if( argc < 3 )
    {
	throw HeyParserException("Missing required arguments.");
    }

    /* Store the arguments and then */
    this->hp_ArgCount = argc;
    this->hp_ArgValue = argv;
    this->hp_ArgIndex = 3;

    /* ... verify the server name, */
    if( strlen(argv[1]) == 0 )
    {
	throw HeyParserException("Empty server name");
    }

    /* ... figure out the 'what' value, and */
    if( strcasecmp(argv[2], "list") == 0 )
    {
	this->hp_What = LIST_PROPERTIES;
    }
    else if( strcasecmp(argv[2], "get") == 0 )
    {
	this->hp_What = GET_PROPERTY;
    }
    else if( strcasecmp(argv[2], "set") == 0 )
    {
	this->hp_What = SET_PROPERTY;
    }
    else
    {
	throw HeyParserException("Unknown verb value.");
    }

    /* ... push any property lists. */
    while( (this->hp_ArgIndex < this->hp_ArgCount) &&
	   (strcmp(this->hp_ArgValue[this->hp_ArgIndex], "to") != 0) )
    {
	struct Pair pair;

	this->nextProperty(pair.p_Name, pair.p_Value);
	this->hp_PropertyStack.push(pair);
    }
}

HeyParser::~HeyParser()
{
    this->hp_ArgCount = 0;
    this->hp_ArgValue = NULL;
    this->hp_ArgIndex = -1;
}

const char *HeyParser::who(void)
{
    const char *retval;

    retval = this->hp_ArgValue[1];
    
    return( retval );
}

HeyParser::hp_what_t HeyParser::what(void)
{
    return( this->hp_What );
}

const char *HeyParser::consumeArg()
    throw (HeyParserException)
{
    const char *retval;

    if( this->hp_ArgIndex >= this->hp_ArgCount )
    {
	throw HeyParserException();
    }
    
    retval = this->hp_ArgValue[this->hp_ArgIndex];
    this->hp_ArgIndex += 1;

    return( retval );
}

void HeyParser::nextProperty(const char *&name_out, const char *&value_out)
    throw (HeyParserException)
{
    const char *arg;
    
    name_out = NULL;
    value_out = NULL;
    
    if( this->hp_ArgIndex > 3 )
    {
	/* Expect 'of' after the first property is processed. */
	if( strcasecmp(this->consumeArg(), "of") == 0 )
	{
	}
	else
	{
	    throw HeyParserException();
	}
    }

    /* The property name is always required. */
    name_out = this->consumeArg();
    try
    {
	arg = this->consumeArg();
	if( (strcasecmp(arg, "of") == 0) || (strcasecmp(arg, "to") == 0) )
	{
	    /*
	     * Detected a preposition separating properties.  Back up since it
	     * will be consumed on the next go around.
	     */
	    this->hp_ArgIndex -= 1;
	}
	else
	{
	    value_out = arg;
	}
    }
    catch(const HeyParserException &e)
    {
	/* Ignore missing value. */
    }
}

void HeyParser::popProperty(const char *&name_out, const char *&value_out)
    throw (HeyParserException)
{
    struct Pair pair;

    name_out = NULL;
    value_out = NULL;
    
    if( this->hp_PropertyStack.size() == 0 )
    {
	throw HeyParserException();
    }

    pair = this->hp_PropertyStack.top();
    name_out = pair.p_Name;
    value_out = pair.p_Value;
    this->hp_PropertyStack.pop();
}

const char *HeyParser::to(void)
    throw (HeyParserException)
{
    const char *retval;

    /*
     * The processing done in the constructor should leave us on the 'to'
     * argument, if there is one.
     */
    try
    {
	if( strcasecmp(this->consumeArg(), "to") != 0 )
	{
	    throw HeyParserException();
	}
    }
    catch(HeyParserException &e)
    {
	throw HeyParserException("Missing 'to' keyword.");
    }

    try
    {
	retval = this->consumeArg();
    }
    catch(HeyParserException &e)
    {
	throw HeyParserException("Missing 'to' value.");
    }
    
    return( retval );
}
