/*
 * rktimes.c
 *
 * Copyright (c) 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file rktimes.c
 *
 * Main file for a sampling time(1)-like command.  Unlike time(1) though, it
 * will not wait until the end of a run to report the CPU usage, it
 * continuously reports the usage in a form fit for consumption by gnuplot.
 * In addition, it can act as a server for a GKrellm client so the data can be
 * viewed instantaneously in a nice GUI.
 */

#include "config.h"

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/resource.h>

#include <netinet/in.h>

#include <assert_pp.h>

#include "gkemu.h"
#include "childProcess.h"

#include <rk/rk.h>
#include <rk/rk_error.h>

/**
 * Number of times per second to record CPU usage samples.
 */
#define SAMPLES_PER_SECOND 1

/**
 * Maximum number of gkrellm clients.
 */
#define MAX_CLIENTS 8

/**
 * Size of the buffer used to format gkrellm output.
 */
#define CLIENT_BUFFER_MAX 4096

#if !defined(__XSTRING)
/**
 * Convert a macro argument to a string.
 *
 * @param The macro to expand to a string.
 */
#define __XSTRING(x) __STRING(x)
#endif

/**
 * A version of perror(3) that prints out the file and line number.
 *
 * @param x The prefix string for the error string.
 */
#define my_perror(x) perror(x ", file " __FILE__ ", line " __XSTRING(__LINE__))

enum {
    RKTB_DONE,
    RKTB_CREATED_RESOURCE_SET,
};

/*
 * Flags for the rktimes_data.rkt_Flags field.
 *
 * RKTF_DONE - Stop monitoring the resource set/child process.
 * RKTF_CREATED_RESOURCE_SET - rktimes created the resource set for the child.
 */
enum {
    RKTF_DONE = (1L << RKTB_DONE),
    RKTF_CREATED_RESOURCE_SET = (1L << RKTB_CREATED_RESOURCE_SET),
};

/*
 * Global data for the tool.
 *
 * rkt_Flags - Holds the RKTF_ flags.
 * rkt_UtilityName - The name of the utility that is being monitored.
 * rkt_StartTime - The starting time for the child process.
 * rkt_OutputBase - The base name for output files.
 * rkt_Name - The request name of the resource set.
 * rkt_ResourceSet - The child's resource set.
 * rkt_ChildPID - The pid_t of the child process.  This will be set to -1 when
 *                the child dies.
 * rkt_ServerSocket - The server socket listening for gkrellm clients.
 * rkt_ClientBuffer - Work buffer for sending data to gkrellm clients.
 * rkt_ClientSockets - The list of gkrellm clients listening for updates.
 */
static struct {
    unsigned long rkt_Flags;
    char *rkt_UtilityName;
    struct timeval rkt_StartTime;
    char *rkt_OutputBase;
    char *rkt_Name;
    rk_resource_set_t rkt_ResourceSet;
    pid_t rkt_ChildPID;
    
    int rkt_ServerSocket;
    char rkt_ClientBuffer[CLIENT_BUFFER_MAX];
    int rkt_ClientSockets[MAX_CLIENTS];
} rktimes_data;

/**
 * Write a byte array to a client's socket.  If they happen to fall behind or
 * some other error is encountered, the socket is closed immediately and erased
 * from the rktimes_data.rkt_ClientSockets array.
 *
 * @param index The client index in rktimes_data.rkt_ClientSockets.
 * @param buffer The buffer to write to the socket.
 * @param len The length of the buffer.
 */
static void rktClientWrite(unsigned int index, char *buffer, size_t len)
{
    int fd;

    require(index < MAX_CLIENTS);
    require(buffer != NULL);

    if( (fd = rktimes_data.rkt_ClientSockets[index]) >= 0 )
    {
	int rc;

	rc = write(fd, buffer, len);
	if( rc != len )
	{
	    require((rc != -1) || (errno == EPIPE) || (errno == EWOULDBLOCK));

	    /* No mercy! */
	    close(fd);
	    rktimes_data.rkt_ClientSockets[index] = -1;
	}
    }
}

/**
 * A signal handler that passes the received signal on to the child process.
 *
 * @param sig The actual signal number received.
 */
static void sigpass(int sig)
{
    require((sig == SIGINT) || (sig == SIGTERM));
    
    kill(rktimes_data.rkt_ChildPID, sig);
}

/**
 * The SIGALRM signal handler.  This function will be called SAMPLES_PER_SECOND
 * times a second to read the CPU usage of the resource set's processes.  The
 * data is then output to a file and echoed to any gkrellm clients.
 *
 * @param sig The actual signal number received.
 *
 * @sa gkemu.h
 * @sa childProcess.h
 */
static void sigalrm(int sig)
{
    static unsigned long long total_compute, total_nice, total_deadline;
    static unsigned long long total_usage;
#define MAX_PROC_LIST 128
    static pid_t procs[MAX_PROC_LIST];

    unsigned long long inc_usage = 0, total_run, total_idle;
    struct cpu_reserve_attr cra;
    struct timeval curr, run;
    int lpc, len, count;
    rk_reserve_t cr;

    require(sig == SIGALRM);

    /* Get the current time and compute the processes total running time. */
    gettimeofday(&curr, NULL);
    timersub(&curr, &rktimes_data.rkt_StartTime, &run);
    total_run = (run.tv_sec * 1000000) + run.tv_usec;

    /* Get the list of processes in the resource set. */
    if( (count = rk_resource_set_get_proclist(rktimes_data.rkt_ResourceSet,
					      procs,
					      MAX_PROC_LIST)) == -1 )
    {
	my_perror("rk_resource_set_get_proclist");
    }

    /* Walk through each process in the resource set and update its usage. */
    for( lpc = 0; lpc < count; lpc++ )
    {
	struct cpChildProcess *cp;
	
	if( (cp = cpFindChildProcess(procs[lpc])) == NULL )
	{
	    /* The child is new, create a new object for it. */
	    if( (cp = cpCreateChildProcess(procs[lpc])) != NULL )
	    {
		cpOpenOutput(cp,
			     rktimes_data.rkt_OutputBase,
			     &rktimes_data.rkt_StartTime);
	    }
	}
	if( cp != NULL )
	{
	    inc_usage += cpSampleUsage(cp, &run);
	}
    }

    /* Add the usage since the last time this handler was called. */
    total_usage += inc_usage;

    /* Get the current CPU reservation status. */
    if( (cr = rk_resource_set_get_cpu_rsv(rktimes_data.rkt_ResourceSet)) !=
	NULL_RESERVE)
    {
	if( rk_cpu_reserve_get_attr(cr, &cra) != RK_ERROR )
	{
	    unsigned long long rsv_period, rsv_compute, rsv_nice, rsv_deadline;
	    double multiplier;

	    /* Turn the raw reservation data into a form fit for gkrellm. */
	    rsv_period = (cra.period.tv_sec * 1000000) +
		(cra.period.tv_nsec / 1000);
	    multiplier = (1000000.0 / rsv_period) /
		(double)SAMPLES_PER_SECOND;
	    rsv_compute = ((cra.compute_time.tv_sec * 1000000) +
			   (cra.compute_time.tv_nsec / 1000)) * multiplier;
	    if( inc_usage > rsv_compute )
	    {
		rsv_nice = 0;
	    }
	    else
	    {
		rsv_nice = rsv_compute - inc_usage;
		rsv_compute = inc_usage;
	    }
	    rsv_deadline = ((cra.deadline.tv_sec * 1000000) +
			    (cra.deadline.tv_nsec / 1000)) * multiplier;

	    total_deadline += rsv_deadline - (rsv_compute + rsv_nice);
	    total_compute += rsv_compute;
	    total_nice += rsv_nice;
	}
    }
    
    total_idle = total_run - total_usage;

    /* Generate the new update text. */
    len = gkFormatUpdate(rktimes_data.rkt_ClientBuffer, CLIENT_BUFFER_MAX,
			 GKA_CPUUser, total_usage,
			 GKA_CPUIdle, total_idle,
			 GKA_CPUReserveUser, total_compute,
			 GKA_CPUReserveNice, total_nice,
			 GKA_CPUReserveIdle, total_deadline,
			 GKA_Processes, count,
			 GKA_ProcessesRunning, 0,
			 GKA_UpTime, &run,
			 GKA_TAG_DONE);

    /* Broadcast the update to the clients. */
    for( lpc = 0; lpc < MAX_CLIENTS; lpc++ )
    {
	if( rktimes_data.rkt_ClientSockets[lpc] != -1 )
	{
	    rktClientWrite(lpc, rktimes_data.rkt_ClientBuffer, len);
	}
    }
}

/**
 * Handle a SIGCHLD signal.  This function will be called when the child exits.
 *
 * @param The actual signal number received.
 */
static void sigchld(int sig)
{
    require(sig == SIGCHLD);
    
    rktimes_data.rkt_Flags |= RKTF_DONE;
}

/**
 * Handle a SIGINT/SIGTERM signal when no utility is being monitored.
 *
 * @param The actual signal number received.
 */
static void sigexit(int sig)
{
    require((sig == SIGINT) || (sig == SIGTERM));
    
    rktimes_data.rkt_Flags |= RKTF_DONE;
}

/**
 * Handle a SIGIO signal.  This function will be called when the server socket
 * has a new connection waiting to be accept(2)'d.
 *
 * @param The actual signal number received.
 */
static void sigio(int sig)
{
    struct sockaddr saddr;
    socklen_t saddrlen;
    int fd;

    require(sig == SIGIO);

    saddrlen = sizeof(saddr);
    if( (fd = accept(rktimes_data.rkt_ServerSocket, &saddr, &saddrlen)) >= 0 )
    {
	int fl, lpc;

	/*
	 * Set the client socket to be non-blocking, we do not want to get
	 * stuck waiting for someone.
	 */
	fl = fcntl(fd, F_GETFL, 0);
	fcntl(fd, F_SETFL, fl | O_NONBLOCK);
	for( lpc = 0;
	     (lpc < MAX_CLIENTS) &&
		 (rktimes_data.rkt_ClientSockets[lpc] != -1);
	     lpc++ );
	if( lpc < MAX_CLIENTS )
	{
	    int len;

	    rktimes_data.rkt_ClientSockets[lpc] = fd;

	    /* Send the preamble and */
	    len = gkFormatPreamble(
		rktimes_data.rkt_ClientBuffer,
		CLIENT_BUFFER_MAX,
		GKA_HostName, rktimes_data.rkt_Name,
		GKA_SystemName, rktimes_data.rkt_UtilityName,
		GKA_TAG_DONE);
	    rktClientWrite(lpc, rktimes_data.rkt_ClientBuffer, len);

	    /* ... the initial update. */
	    len = gkFormatInitialUpdate(
		rktimes_data.rkt_ClientBuffer,
		CLIENT_BUFFER_MAX,
		GKA_TAG_DONE);
	    rktClientWrite(lpc, rktimes_data.rkt_ClientBuffer, len);
	}
	else
	{
	    /* Too many clients... */
	    close(fd);
	}
    }
    else
    {
	my_perror("accept");
    }
}

/**
 * Create a resource set and CPU reserve for ourself.  Otherwise, we might not
 * get CPU time to do the recording.
 *
 * @param name The name for the resource set.
 * @return A resource set for this process.
 */
static rk_resource_set_t rktCreateSelfResourceSet(char *name)
{
    rk_resource_set_t rs, retval = NULL_RESOURCE_SET;

    require(name != NULL);
    require(strlen(name) > 0);
    require(strlen(name) < RK_NAME_LEN);
    
    if( (rs = rk_resource_set_create(name)) != NULL )
    {
	struct cpu_reserve_attr cra;
	rk_reserve_t cr;
	
	memset(&cra, 0, sizeof(cra));
	/* XXX Come up with some better numbers for the reserve. */
	cra.compute_time.tv_sec = 0;
	cra.compute_time.tv_nsec = 5000000;
	cra.period.tv_sec = 0;
	cra.period.tv_nsec = 100000000;
	cra.deadline = cra.period;
	cra.blocking_time = cra.start_time = (struct timespec){ 0, 0 };
	cra.reserve_type.sch_mode = RSV_SOFT;
	cra.reserve_type.enf_mode = RSV_SOFT;
	cra.reserve_type.rep_mode = RSV_SOFT;
	cra.processor = RK_ANY_CPU;
	if( rk_cpu_reserve_create(rs, &cr, &cra) != RK_ERROR )
	{
	    /*
	     * Note that we do not attach the process here since any children
	     * we fork(2) off would inherit the resource set.
	     */
	    retval = rs;
	}
	else
	{
	    my_perror("rk_cpu_reserve_create");
	    rk_resource_set_destroy(rs);
	}
    }
    else
    {
	my_perror("rk_resource_set_create");
    }
    return( retval );
}

/**
 * Get a resource set for the child process.  If a resource set already exists
 * with the requested name, that will be used.  Otherwise, a new one will be
 * created.  If no name is requested, a new set will be created with the
 * default name.
 *
 * @param requested_name The requested resource set name or NULL if the default
 * should be used.
 * @param default_name The default resource set name to use if requested_name
 * is NULL.
 * @return The resource set to use for the child process or NULL_RESOURCE_SET
 * if a set could not be created.
 */
static rk_resource_set_t rktGetChildResourceSet(char *requested_name,
						char *default_name)
{
    rk_resource_set_t retval = NULL_RESOURCE_SET;

    require((requested_name == NULL) ||
	    ((strlen(requested_name) > 0) &&
	     (strlen(requested_name) < RK_NAME_LEN)));
    require(default_name != NULL);
    require(strlen(default_name) > 0);
    require(strlen(default_name) < RK_NAME_LEN);

    /* First, try to find an existing resource set. */
    if( requested_name != NULL )
    {
#define MAX_RK_SETS 128
	rk_resource_set_t rk_sets[MAX_RK_SETS];
	char rs_name[RK_NAME_LEN + 1];
	int lpc, actual;
	
	actual = rk_resource_sets_get_list(rk_sets, MAX_RK_SETS);
	for( lpc = 0; (lpc < actual) && (retval == NULL_RESOURCE_SET); lpc++ )
	{
	    rs_name[0] = '\0';
	    if( rk_resource_set_get_name(rk_sets[lpc], rs_name) == 0 )
	    {
		if( strcmp(rs_name, rktimes_data.rkt_Name) == 0 )
		{
		    retval = rk_sets[lpc];
		}
	    }
	    else
	    {
		my_perror("rk_resource_set_get_name");
	    }
	}
#undef MAX_RK_SETS
    }
    /* Next, try to make a new set if one was not found. */
    if( retval == NULL_RESOURCE_SET )
    {
	char *actual_name;
	
	if( requested_name != NULL )
	{
	    actual_name = requested_name;
	}
	else
	{
	    actual_name = default_name;
	}
	if( (retval = rk_resource_set_create(actual_name)) != NULL )
	{
	    rktimes_data.rkt_Flags |= RKTF_CREATED_RESOURCE_SET;
	}
	else
	{
	    my_perror("rk_resource_set_create");
	}
    }
    return( retval );
}

/**
 * Print out the usage statement to a file handle.
 *
 * @param file The destination file handle for the usage output.
 * @param prog_name The program name.
 */
static void rktUsage(FILE *file, char *prog_name)
{
    require(file != NULL);
    require(prog_name != NULL);
    require(strlen(prog_name) > 0);
    
    fprintf(file,
	    "Usage: %s [options] -- <command> [argument ...]\n",
	    prog_name);
    fprintf(file,
	    "\n"
	    "Create a resource set and periodically record the CPU\n"
	    "usage of the attached process and its children.\n"
	    "\n"
	    "Options:\n"
	    "\t-h\t\tThis help message.\n"
	    
	    "\t-o <file>\tBase name for the output files. (Default: rktimes)\n"
	    
	    "\t-n <name>\tName of the RK resource set.  If a resource set\n"
	    "\t\t\twith that name already exists, that will be used.\n"
	    "\t\t\tOtherwise, a new one with that name will be created.\n"
	    "\t\t\t(Default: rktimes)\n"
	    
	    "\t-p <port>\tServer port for GKrellmd emulation.\n");
}

/**
 * Process the command line options.
 *
 * @param argc_inout Pointer to main's argc variable.  On return, the variable
 * will contain the number of arguments remaining after option processing.
 * @param argv_inout Pointer to main's argv variable.  On return, the variable
 * will contain the remaining argument values.
 * @return True if the options were processed correctly, false otherwise.
 */
static int rktProcessOptions(int *argc_inout, char **argv_inout[])
{
    int ch, retval = 1;
    char *prog_name;
    char **argv;
    int argc;

    require(argc_inout != NULL);
    require(argv_inout != NULL);

    argc = *argc_inout;
    argv = *argv_inout;
    prog_name = argv[0];
    while( ((ch = getopt(argc, argv, "ho:n:p:")) != -1) && (retval == 1) )
    {
	switch( ch )
	{
	case 'o':
	    /* Output file base name. */
	    rktimes_data.rkt_OutputBase = optarg;
	    break;
	case 'n':
	    /* Resource set name. */
	    if( strlen(optarg) > RK_NAME_LEN )
	    {
		fprintf(stderr,
			"%s: -n value too long (only %d characters allowed)\n",
			prog_name,
			RK_NAME_LEN);
		retval = 0;
	    }
	    else
	    {
		rktimes_data.rkt_Name = optarg;
	    }
	    break;
	case 'p':
	    /* Server port number for GKrellm emulation. */
	    {
		int port;
		
		if( (sscanf(optarg, "%d", &port) == 1) &&
		    (port > 0) && (port < 65536) )
		{
		    int fd;
		    
		    if( (fd = socket(PF_INET, SOCK_STREAM, 0)) >= 0 )
		    {
			struct sockaddr_in sin;
			int on = 1;
			
			(void)setsockopt(fd,
					 SOL_SOCKET,
					 SO_REUSEADDR,
					 (char *)&on,
					 sizeof(on));
#if defined(BSD44)
			sin.sin_len = sizeof(sin);
#endif
			sin.sin_family = AF_INET;
			sin.sin_port = htons(port);
			sin.sin_addr.s_addr = INADDR_ANY;
			if( bind(fd,
				 (struct sockaddr *)&sin,
				 sizeof(sin)) == 0 )
			{
			    if( listen(fd, 5) == 0 )
			    {
				int fl;

				fl = fcntl(fd, F_GETFL, 0);
				fl |= O_NONBLOCK |
#if defined(O_ASYNC)
				    O_ASYNC
#elif defined(FASYNC)
				    FASYNC
#endif
				    ;
				fcntl(fd, F_SETFL, fl);
				fcntl(fd, F_SETOWN, getpid());
			    }
			    else
			    {
				my_perror("listen");
				retval = 0;
			    }
			}
			else
			{
			    my_perror("bind");
			    retval = 0;
			}
		    }
		    else
		    {
			my_perror("socket");
			retval = 0;
		    }
		    if( retval )
		    {
			rktimes_data.rkt_ServerSocket = fd;
		    }
		    else
		    {
			close(fd);
		    }
		}
		else
		{
		    fprintf(stderr,
			    "%s: Invalid -p value: %s\n",
			    prog_name,
			    optarg);
		    retval = 0;
		}
	    }
	    break;
	case 'h':
	case '?':
	default:
	    retval = 0;
	    break;
	}
    }
    *argc_inout -= optind;
    *argv_inout += optind;
    /* Optionally skip the '--' that is used to terminate the option list. */
    if( (*argc_inout > 0) && (strcmp(*argv_inout[0], "--") == 0) )
    {
	argc_inout -= 1;
	argv_inout += 1;
    }
    return( retval );
}

/**
 * The parent portion of the fork(2) between rktimes and the monitored utility.
 * This function will attach itself to a resource set to ensure that it has
 * some CPU time to work and then wait for gkrellm connections or the child's
 * death.
 *
 * @param rs The resource set for the rktimes process.
 * @return The return code for main().
 *
 * @sa rktChildPart
 */
static int rktParentPart(rk_resource_set_t rs)
{
    int retval = EXIT_FAILURE;
    struct itimerval itv;
    struct sigaction sa;
    sigset_t sigmask;

    require(rs != NULL_RESOURCE_SET);
    
    /* Attach ourselves here to avoid inheritance problem. */
    if( rk_resource_set_attach_process(rs, getpid()) < 0 )
    {
	my_perror("rk_resource_set_attach_process");
	retval = EXIT_FAILURE;
    }

    /*
     * We use signals as a primitive event system, so any that we use and are
     * not passed on to the child need to be blocked before setting up the
     * handlers.  Once everything has been setup, we will use sigsuspend(2) to
     * atomically unblock and wait for the signals to arrive.
     */
    sigemptyset(&sigmask);
    sigaddset(&sigmask, SIGALRM);
    sigaddset(&sigmask, SIGCHLD);
    sigaddset(&sigmask, SIGIO);
    if( sigprocmask(SIG_BLOCK, &sigmask, NULL) < 0 )
    {
	my_perror("sigprocmask");
    }

    /* Setup the signal handlers. */
    sa.sa_mask = sigmask;
    sa.sa_flags = 0;
#if defined(SA_RESTART)
    sa.sa_flags |= SA_RESTART;
#endif

    signal(SIGPIPE, SIG_IGN);

    sa.sa_handler = sigalrm;
    sigaction(SIGALRM, &sa, NULL);

    sa.sa_handler = sigio;
    sigaction(SIGIO, &sa, NULL);

    if( rktimes_data.rkt_ChildPID == 0 )
    {
	/*
	 * No utility is being monitored, just a resource set.  Therefore we
	 * need to exit when these signals are received since there is no
	 * child to pass them to.
	 */
	sa.sa_handler = sigexit;
	sigaction(SIGINT, &sa, NULL);
	sigaction(SIGTERM, &sa, NULL);
    }
    else
    {
	/* Catch SIGCHLD so we exit with the child. */
	sa.sa_handler = sigchld;
	sigaction(SIGCHLD, &sa, NULL);

	/*
	 * Pass SIGINT/SIGTERM onto the child so they can handle them as they
	 * choose.  If they exit, we will get the SIGCHLD and exit shortly
	 * thereafter.
	 */
	sa.sa_handler = sigpass;
	sigaction(SIGINT, &sa, NULL);
	sigaction(SIGTERM, &sa, NULL);
    }
    
    itv.it_interval.tv_sec = 0;
    itv.it_interval.tv_usec = 1000000 / SAMPLES_PER_SECOND;
    itv.it_value.tv_sec = 0;
    itv.it_value.tv_usec = 1000000 / SAMPLES_PER_SECOND;
    /* XXX Use rk periodic stuff here. */
    if( setitimer(ITIMER_REAL, &itv, NULL) == 0 )
    {
	sigset_t empty_sigmask;
	int status;

	sigemptyset(&empty_sigmask);
	
	/* Keep handling signals until the child has died. */
	while( !(rktimes_data.rkt_Flags & RKTF_DONE) )
	{
	    sigsuspend(&empty_sigmask);
	}

	/* Cleanup the child's status. */
	if( rktimes_data.rkt_ChildPID == 0 )
	{
	    /*
	     * No child was being monitored, ergo, there is nothing to wait for
	     */
	}
	else if( wait(&status) >= 0 )
	{
	    if( WIFEXITED(status) )
	    {
		retval = WEXITSTATUS(status);
	    }
	    else if( WIFSIGNALED(status) )
	    {
		retval = EXIT_SUCCESS;
	    }
	    else
	    {
		retval = EXIT_FAILURE;
	    }
	}
	else
	{
	    my_perror("wait");
	    retval = EXIT_FAILURE;
	}
    }
    else
    {
	my_perror("setitimer");
	retval = EXIT_FAILURE;
    }

    return( retval );
}

/**
 * The child portion of the fork(2) between rktimes and the monitored utility.
 * This function will attach the child to the given resource set and exec(3)
 * the utility to monitor with the given arguments.
 *
 * @param rs The child's resource set.
 * @param argv The utility to startup and its arguments.
 * @return A failure exit code, otherwise this function will not return because
 * of the exec(3).
 *
 * @sa rktParentPart
 */
static int rktChildPart(rk_resource_set_t rs, char *argv[])
{
    int retval = EXIT_SUCCESS;
    
    /*
     * Make sure any children created by the monitored utility
     * are also attached to the resource set.
     */
    rk_inherit_mode(1);
    if( rk_resource_set_attach_process(rs, getpid()) == 0 )
    {
    }
    else
    {
	my_perror("rk_resource_set_attach_process");
	retval = EXIT_FAILURE;
    }
    
    if( retval == EXIT_SUCCESS )
    {
	execvp(argv[0], argv);

	/* FALLTHROUGH, normal operation will not reach this point. */
	perror(argv[0]);
	switch( errno )
	{
	case ENOENT:
	case EPERM:
	    retval = 127;
	    break;
	default:
	    retval = EXIT_FAILURE;
	    break;
	}
    }
    return( retval );
}

int main(int argc, char *argv[])
{
    int lpc, retval = EXIT_FAILURE;

    /* Default values. */
    rktimes_data.rkt_OutputBase = "rktimes";
    rktimes_data.rkt_ServerSocket = -1;
    for( lpc = 0; lpc < MAX_CLIENTS; lpc++ )
    {
	rktimes_data.rkt_ClientSockets[lpc] = -1;
    }

    /* Initialize the internal bits first. */
    if( cpInitChildProcessData() )
    {
	char *prog_name = argv[0];
	int rc;

	rc = rktProcessOptions(&argc, &argv);
	if( rc && ((argc > 0) || (rktimes_data.rkt_Name != NULL)) )
	{
	    char self_rs_name[RK_NAME_LEN + 1];
	    rk_resource_set_t rs;

	    if( argc > 0 )
		rktimes_data.rkt_UtilityName = argv[0];
	    else
		rktimes_data.rkt_UtilityName = "<none>";
	    snprintf(self_rs_name, RK_NAME_LEN + 1, "rktimes.%d", getpid());
	    rs = rktCreateSelfResourceSet(self_rs_name);
	    rktimes_data.rkt_ResourceSet =
		rktGetChildResourceSet(rktimes_data.rkt_Name, "rktimes");
	    if( (rs != NULL_RESOURCE_SET) &&
		(rktimes_data.rkt_ResourceSet != NULL_RESOURCE_SET) )
	    {
		gettimeofday(&rktimes_data.rkt_StartTime, NULL);
		if( argc == 0 )
		{
		    retval = rktParentPart(rs);
		}
		else if( (rktimes_data.rkt_ChildPID = fork()) > 0 )
		{
		    retval = rktParentPart(rs);
		}
		else if( rktimes_data.rkt_ChildPID == 0 )
		{
		    return( rktChildPart(rktimes_data.rkt_ResourceSet, argv) );
		}
		else
		{
		    my_perror("fork");
		    retval = EXIT_FAILURE;
		}
	    }
	    if( rktimes_data.rkt_Flags & RKTF_CREATED_RESOURCE_SET )
	    {
		rk_resource_set_destroy(rktimes_data.rkt_ResourceSet);
	    }
	    rk_resource_set_destroy(rs);
	}
	else
	{
	    if( rc && (rktimes_data.rkt_Name == NULL) )
	    {
		fprintf(stderr, "%s: Missing utility to monitor\n", prog_name);
	    }
	    rktUsage(stderr, prog_name);
	}
	cpKillChildProcessData();
    }
    
    return( retval );
}
