/*
 * rt_server_hourglass.cc
 *
 * Copyright (c) 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file rt_server_hourglass.cc
 */

#include "config.h"

#include <signal.h>

#include <iostream>

#include <BrokerC.h>
#include <BasicTaskKernelImpl.h>

#include "RTServer_busy.hh"
#include "RTServerWrapper.hh"

static CORBA::ORB_var orb;
static RTServer_busy *rtsi;

static void sigalrm_handler(int sig)
{
    orb->shutdown();

    delete rtsi;
    
    exit(0);
}

int main(int argc, char *argv[])
{
    int retval = EXIT_FAILURE;

#if defined(HAVE_ACE)
    /* Put a sock in ACE */
    ACE_Log_Msg::instance()->priority_mask(0, ACE_Log_Msg::PROCESS);
#endif

    try
    {
	orb = CORBA::ORB_init(argc, argv);

	CORBA::Object_var obj = orb->resolve_initial_references("RootPOA");

	PortableServer::POA_var root_poa =
	    PortableServer::POA::_narrow(obj.in());
	PortableServer::POAManager_var mgr = root_poa->the_POAManager();
	mgr->activate();

	/* QuoKernel setup */
	quo::QuoKernel_var qk;
	
	try
	{
	    /* ... check for an IOR from the Java kernel. */
	    obj = orb->string_to_object("file://quoKernel.ior");
	    qk = quo::QuoKernel::_narrow(obj.in());
	    if( CORBA::is_nil(qk.in()) || qk->_non_existent() )
	    {
		qk = quo::QuoKernel::_nil();
	    }
	}
	catch(const CORBA::SystemException &e)
	{
	    qk = quo::QuoKernel::_nil();
	}
	if( CORBA::is_nil(qk.in()) )
	{
	    Qosket::Basic::BasicTaskKernelImpl *btki;
	    
	    /* ... no joy on Java kernel, create our own. */
	    btki = new Qosket::Basic::BasicTaskKernelImpl(orb);
	    qk = btki->_this();
	}

	const char *manager_ior = "file://manager.ior";
	const char *obj_name = "RTServer";
	const char *obj_ior = "rtserver.ior";
	CORBA::ULong period = 10000;
	struct {
	    CORBA::ULong data[128];
	    CORBA::ULong length;
	} compute = { { 0 }, 0 };
	const char *tf_ior = NULL;
	unsigned long run_time = 0;

	if( (argc % 2) == 1 )
	{
	    Broker::RealTimeAdvocate_var rta;
	    Broker::Manager_var manager;
	    int lpc;

	    for( lpc = 1; lpc < argc; lpc += 2 )
	    {
		if( strcmp(argv[lpc], "-name") == 0 )
		{
		    obj_name = argv[lpc + 1];
		}
		else if( strcmp(argv[lpc], "-manager") == 0 )
		{
		    manager_ior = argv[lpc + 1];
		}
		else if( strcmp(argv[lpc], "-task-factory") == 0 )
		{
		    tf_ior = argv[lpc + 1];
		}
		else if( strcmp(argv[lpc], "-period") == 0 )
		{
		    if( sscanf(argv[lpc + 1], "%d", &period) != 1 )
		    {
			cerr << "Period value is not a number: "
			     << argv[lpc + 1]
			     << endl;
			throw CORBA::BAD_PARAM();
		    }
		}
		else if( strcmp(argv[lpc], "-compute") == 0 )
		{
		    if( sscanf(argv[lpc + 1],
			       "%d",
			       &compute.data[compute.length]) == 1 )
		    {
			compute.length += 1;
		    }
		    else
		    {
			cerr << "Compute value is not a number: "
			     << argv[lpc + 1]
			     << endl;
			throw CORBA::BAD_PARAM();
		    }
		}
		else if( strcmp(argv[lpc], "-o") == 0 )
		{
		    obj_ior = argv[lpc + 1];
		}
		else if( strcmp(argv[lpc], "-for") == 0 )
		{
		    if( sscanf(argv[lpc + 1], "%d", &run_time) == 0 )
		    {
			cerr << "'-for' value is not an integer: "
			     << argv[lpc + 1]
			     << endl;
		    }
		}
		else
		{
		    cerr << "Ignoring unknown arg: "
			 << argv[lpc]
			 << endl;
		}
	    }
	    
	    if( run_time )
	    {
		struct itimerval itv;
		
		signal(SIGALRM, sigalrm_handler);
		itv.it_interval.tv_sec = 0;
		itv.it_interval.tv_usec = 0;
		itv.it_value.tv_sec = run_time;
		itv.it_value.tv_usec = 0;
		if( setitimer(ITIMER_REAL, &itv, NULL) < 0 )
		{
		    perror("setitimer");
		}
	    }
	    
	    if( compute.length == 0 )
	    {
		compute.data[0] = 1000;
		compute.length = 1;
	    }

	    if( (manager_ior != NULL) && (strlen(manager_ior) > 0) )
	    {
		obj = orb->string_to_object(manager_ior);
		manager = Broker::Manager::_narrow(obj.in());
		if( CORBA::is_nil(manager.in()) )
		{
		    cerr << "Invalid manager IOR: "
			 << manager_ior
			 << endl;
		    throw CORBA::BAD_PARAM();
		}
	    }

	    if( !CORBA::is_nil(manager.in()) && (tf_ior != NULL) )
	    {
		char filename[1024];
		
		Broker::ScheduleParameters sp;
		Broker::TaskFactory_var tf;
		Broker::TaskParameters tp;
		Broker::Task_var task;

		obj = orb->string_to_object(tf_ior);
		tf = Broker::TaskFactory::_narrow(obj.in());
		if( CORBA::is_nil(tf.in()) )
		{
		    cerr << "Invalid task factory IOR: "
			 << tf_ior
			 << endl;
		    throw CORBA::BAD_PARAM();
		}
		tp.length(6);
		tp[0].name = "name";
		tp[0].value <<= obj_name;
		tp[1].name = "pid";
		tp[1].value <<= (CORBA::Long)getpid();
		tp[2].name = "res log";
		sprintf(filename, "/tmp/%s-res.out", obj_name);
		tp[2].value <<= filename;
		tp[3].name = "cpu log";
		sprintf(filename, "/tmp/%s-cpu.out", obj_name);
		tp[3].value <<= filename;
		tp[4].name = "advice log";
		sprintf(filename, "/tmp/%s-advice.out", obj_name);
		tp[4].value <<= filename;
		tp[5].name = "complete log";
		sprintf(filename, "/tmp/%s-complete.out", obj_name);
		tp[5].value <<= filename;
		try
		{
		    task = tf->CreateTask(tp);
		    rta = Broker::RealTimeAdvocate::_narrow(task.in());
		    if( CORBA::is_nil(rta.in()) )
		    {
			cerr << "Unable to make RealTimeAdvocate..."
			     << endl;
			throw CORBA::UNKNOWN();
		    }
		    sp.length(1);
		    sp[0].name = "period";
		    sp[0].value <<= period;
		    rta->BeginCPUScheduling(manager.in(), sp);
		}
		catch(const Broker::DuplicateTaskParameter &e)
		{
		    cerr << e << endl;
		    cerr << e.name << endl;
		}
		catch(const Broker::InvalidTaskParameter &e)
		{
		    cerr << e << endl;
		    cerr << e.message << endl;
		}
		catch(const Broker::MissingTaskParameter &e)
		{
		    cerr << e << endl;
		    cerr << e.name << endl;
		}
	    }

	    rtsi = new RTServer_busy(period,
				     &compute.data[0],
				     compute.length);
	    RTServer_var real_rts;
	    
	    real_rts = rtsi->_this();

	    RTServerWrapper *rtsw = new RTServerWrapper(rta.in(),
							period,
							period);
	    RTServer_var rts;
	    
	    rts = rtsw->_this();
	    
	    rtsw->initSysconds(qk);
	    rtsw->initCallbacks(orb);
	    rtsw->linkContract(qk);
	    rtsw->linkRemoteObject(real_rts.in());

	    if( obj_ior != NULL )
	    {
		ofstream ostr(obj_ior);
		ostr << orb->object_to_string(rts.in()) << endl;
		ostr.close();
	    }

	    orb->run();

	    if( !CORBA::is_nil(rta.in()) )
	    {
		rta->EndCPUScheduling();
	    }
		
	    orb->shutdown();

	    retval = EXIT_SUCCESS;
	}
	else
	{
	    cerr << "Usage: " << argv[0] << endl

		 << "\t[-name <name>]"
		 << "\t\tThe name for the object in the NamingService. "
		 << "(default: " << obj_name << ")" << endl

		 << "\t[-period <usecs>]"
		 << "\tThe period in microseconds. "
		 << "(default: " << period << ")" << endl

		 << "\t[-compute <usecs>]"
		 << "\tThe compute time per period in microseconds. "
		 << "(default: " << period << ")" << endl

		 << "\t[-manager <ior>]"
		 << "\tThe IOR of the manager. "
		 << "(default: " << manager_ior << ")" << endl

		 << "\t[-task-factory <ior>]"
		 << "\tThe IOR of the task factory. "
		 << "(default: "
		 << (tf_ior == NULL ? "<none>" : tf_ior)
		 << ")" << endl

		 << "\t[-o <ior file>]"
		 << "\t\tThe IOR output file. "
		 << "(default: "
		 << (obj_ior == NULL ? "<none>" : obj_ior)
		 << ")" << endl;
	}
    }
    catch(const CORBA::SystemException &e)
    {
	cerr << "Caught Exception: " << e << endl;
    }
    catch(...)
    {
	cerr << "Caught an unhandled exception" << endl;
    }
    return( retval );
}
