/*
 * all_test1.cc
 *
 * Copyright (c) 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file all_test1.cc
 *
 * An all-up test of the CPU broker components.
 */

#include "config.h"

#if !defined(DEBUG)
#define DEBUG 1
#endif

#include <stdlib.h>

#include <assert.h>
#include <assert_pp.h>

#include <timesFile.h>

#include <rk.h>
#include <rk_stub.h>

#include <iostream>

#include <BrokerC.h>
#include <StrictPolicyC.h>

#include <ManagerImpl.hh>
#include <BasicDelegate.hh>

#include <StrictPolicyImpl.hh>

#include <ExactTaskFactory.hh>
#include <MaxTaskFactory.hh>
#include <GlacialTaskFactory.hh>

#if defined(HAVE_QUO)
#include <ContractDelegate.hh>
#include <BasicTaskKernelImpl.h>
#include <BasicTaskQosketImpl.hh>
#include <ExactTaskFactory_Quo.hh>
#endif

int main(int argc, char *argv[])
{
    int retval = EXIT_SUCCESS;

    /* Initialize the simulator */
    rk_stub_set_mode(RK_STUB_SIM);

#if defined(HAVE_ACE)
    /* Put a sock in ACE */
    ACE_Log_Msg::instance()->priority_mask(0, ACE_Log_Msg::PROCESS);
#endif

    /* ORB/POA setup */
    CORBA::ORB_var orb = CORBA::ORB_init(argc, argv);
    CORBA::Object_var obj = orb->resolve_initial_references("RootPOA");
    
    PortableServer::POA_var root_poa = PortableServer::POA::_narrow(obj.in());
    PortableServer::POAManager_var mgr = root_poa->the_POAManager();
    mgr->activate();

    /* Create some input files for each process to be simulated. */
    assert(tfMakeTimesFile("c1_cpu_times",
			   (const float[]){
	55.0, /*
	       * repeatedly require 55% of the CPU for task one to meet the
	       * deadline
	       */
	/* Add more values here to change the sequence of CPU requirements */
	TIMES_FILE_TERMINATOR
    }));
    assert(tfMakeTimesFile("c2_cpu_times",
			   (const float[]){
	20.0, /* ... 20% for task two */
	TIMES_FILE_TERMINATOR
    }));
    assert(tfMakeTimesFile("c3_cpu_times",
			   (const float[]){
	20.0, /* ... 20% for task three */
	TIMES_FILE_TERMINATOR
    }));

#if defined(HAVE_QUO)
    /* QuoKernel setup */
    quo::QuoKernel_var qk;

    try
    {
	/* ... check for an IOR from the Java kernel. */
	obj = orb->string_to_object("file://quoKernel.ior");
	qk = quo::QuoKernel::_narrow(obj.in());
	if( CORBA::is_nil(qk.in()) || qk->_non_existent() )
	{
	    qk = quo::QuoKernel::_nil();
	}
    }
    catch(const CORBA::SystemException &e)
    {
	qk = quo::QuoKernel::_nil();
    }
    if( CORBA::is_nil(qk.in()) )
    {
	Qosket::Basic::BasicTaskKernelImpl *btki;

	/* ... no joy on Java kernel, create our own. */
	btki = new Qosket::Basic::BasicTaskKernelImpl(orb);
	qk = btki->_this();
    }
#endif

    /* Broker setup */
    Broker::Manager_var manager;
    ManagerImpl *mi;

    mi = new ManagerImpl("dummy");
    manager = mi->_this();

    StrictPolicy_var policy;
    StrictPolicyImpl *spi;

    spi = new StrictPolicyImpl("StrictPolicy");
    policy = spi->_this();

    mi->CurrentPolicy(policy.in());

#if defined(HAVE_QUO)
    /* Make a qosket for the first task. */
    BasicTaskQosketImpl *btqi;
    quo::Contract_var contract;
    quo::ValueSC_var last_status;

    btqi = new BasicTaskQosketImpl();
    btqi->initSysconds(qk);
    btqi->initCallbacks(orb);
    contract = btqi->initContract(qk);

    last_status = btqi->LastStatus();
    
    ContractDelegate c1(contract.in(), "c1", 10000, 10000);
    {
	int slot;

	/*
	 * The Delegate created some of the TaskParameters, but we need to add
	 * any needed by the ExactTaskAdvocate_Quo.
	 */
	slot = c1.bd_TaskDescription.length();
	c1.bd_TaskDescription.length(slot + 1);
	c1.bd_TaskDescription[slot].name = "last_status";
	c1.bd_TaskDescription[slot].value <<= last_status.in();
    }
#else
    BasicDelegate c1("c1", 10000, 10000);
#endif

    /* Make the delegates and, in turn, the fake processes. */
    BasicDelegate c2("c2", 20000, 20000);
    BasicDelegate c3("c3", 10000, 10000);

    /* Make the tasks/advocates */
    Broker::TaskFactory_var tf;
    Broker::Task_var task;

    ExactTaskFactory *etf;

    etf = new ExactTaskFactory("ExactTaskFactory");

    tf = etf->_this();
    
    /* ... for process one */
#if defined(HAVE_QUO)
    ExactTaskFactory_Quo *etfq;

    etfq = new ExactTaskFactory_Quo("ExactTaskFactory_Quo");

    tf = etfq->_this();
#else
    tf = etf->_this();
#endif

    task = tf->CreateTask(c1.bd_TaskDescription);
    c1.bd_Advocate = Broker::RealTimeAdvocate::_narrow(task.in());
    assert(!CORBA::is_nil(c1.bd_Advocate.in()));
    c1.bd_Advocate->BeginCPUScheduling(manager.in(), c1.bd_CPUSchedule);

    /* ... for process two */
    tf = etf->_this();

    task = tf->CreateTask(c2.bd_TaskDescription);
    c2.bd_Advocate = Broker::RealTimeAdvocate::_narrow(task.in());
    assert(!CORBA::is_nil(c2.bd_Advocate.in()));
    c2.bd_Advocate->BeginCPUScheduling(manager.in(), c2.bd_CPUSchedule);

    /* ... for process three */
    GlacialTaskFactory *gtf;

    gtf = new GlacialTaskFactory("GlacialTaskFactory");

    tf = gtf->_this();

    task = tf->CreateTask(c3.bd_TaskDescription);
    c3.bd_Advocate = Broker::RealTimeAdvocate::_narrow(task.in());
    assert(!CORBA::is_nil(c3.bd_Advocate.in()));
    c3.bd_Advocate->BeginCPUScheduling(manager.in(), c3.bd_CPUSchedule);

    /* Start the simulation. */
    struct timespec curr, end = { 1, 0 };

    /* ... stage one, all the same priority */
    do {
	rk_stub_next_tick();
	rk_clock_gettime(CLOCK_REALTIME, &curr);
    } while( curr.tv_sec < end.tv_sec );

    /* ... stage two, bump process two's priority */
    policy->SetTaskPriority(c2.bd_Advocate.in(), 1);

    end.tv_sec = 2;
    
    do {
	rk_stub_next_tick();
	rk_clock_gettime(CLOCK_REALTIME, &curr);
    } while( curr.tv_sec < end.tv_sec );

    /* ... stage three, bump process three's priority */
    policy->SetTaskPriority(c3.bd_Advocate.in(), 1);

    end.tv_sec = 3;
    
    do {
	rk_stub_next_tick();
	rk_clock_gettime(CLOCK_REALTIME, &curr);
    } while( curr.tv_sec < end.tv_sec );

    /* ... stage four, bump process one's priority past the others */
    policy->SetTaskPriority(c1.bd_Advocate.in(), 2);

    end.tv_sec = 4;
    
    do {
	rk_stub_next_tick();
	rk_clock_gettime(CLOCK_REALTIME, &curr);
    } while( curr.tv_sec < end.tv_sec );
    
    return( retval );
}
